/*
 * NellyMoser ASAO codec
 * Copyright (C) 2007-2008  UAB "DKD"
 * Copyright (C) 2007-2008  Joseph Artsimovich <jart@users.sourceforge.net>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FFmpeg; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include <iostream>
#include <vector>
#include <stdexcept>
#include <algorithm>
#include <stdio.h>
#include <stddef.h>
#include <string.h>
#include <stdint.h>
#include <stdlib.h>
#include <math.h>
#include <sndfile.h>
#include <assert.h>
#include <string>

struct FlvTag
{
	uint32_t prev_tag_len;
	uint8_t  type;
	uint32_t data_len;
	uint32_t timestamp;
};

// 24 bytes
struct NCI_CODEC_IO_BLOCK
{
	short*   uncompressed_samples; // 0
	short    num_samples; // 4
	// padding
	uint8_t* compressed_bytes; // 8
	short    num_bytes; // 0xC
	uint16_t decomp_byte_offset; // 0xE
	uint8_t  decomp_bit_offset;  // 0x10
	// padding
	uint16_t compr_byte_offset; // 0x12
	uint8_t  compr_bit_offset; // 0x14
	// padding
};

struct BTYPE
{
	unsigned char*  stream; // 0
	unsigned short  stream_size; // 4
	unsigned short* in_byte_offset_ptr; // 8
	unsigned char*  in_bit_offset_ptr; // 0xC
	unsigned short* out_byte_offset_ptr; // 0x10
	unsigned char*  out_bit_offset_ptr; // 0x14
};

static void* decoder = 0;
static void* encoder = 0;

static size_t FLV_TAG_PREV_SIZE = 0x04;
static size_t FLV_TAG_SIZE = FLV_TAG_PREV_SIZE + 11;
static uint8_t AUDIO_TAG = 0x08;

struct NciContext
{
	uint32_t signature1; // 0xface
	uint32_t unknown1[2];
	uint8_t signature2; // 0x0b
	void* handle; // offset 0x10
	uint32_t unknown2[6];
};

static int initDecode(void** handle_ptr)
{
	*handle_ptr = malloc(256);
	if (!*handle_ptr) {
		return -1;
	}
	memset(*handle_ptr, 0, 256);
	return 0;
}

static void freeDecode(void** handle_ptr)
{
	if (!*handle_ptr) {
		return;	
	}
	
	free(*handle_ptr);
	*handle_ptr = 0;
}

static int initEncode(void** handle_ptr)
{
	*handle_ptr = malloc(256);
	if (!*handle_ptr) {
		return -1;
	}
	memset(*handle_ptr, 0, 256);
	return 0;
}

static void freeEncode(void** handle_ptr)
{
	if (!*handle_ptr) {
		return;	
	}
	
	free(*handle_ptr);
	*handle_ptr = 0;
}

static int nciFreeDecodeACFR16(void** context_ptr)
{
	NciContext* context = *((NciContext **)context_ptr);
	if (!context) {
		return 0x181a;
	}
	
	if (context->signature1 != 0xface) {
		return 0x17d4;
	}
	
	if (context->signature2 != 0x0b) {
		return 0x17de;
	}
	
	context->signature1 = 0;
	context->signature2 = 0;
	freeDecode(&context->handle);
	
	if (!context) {
		return 0;	
	}
	
	free(context);
	*context_ptr = 0;
	return 0;
}

static int nciInitDecodeACFR16(void** context_ptr)
{
	*context_ptr = malloc(44);
	if (!*context_ptr) {
		return 0x1770;
	}
	
	NciContext* context = *((NciContext **)context_ptr);
	context->signature2 = 0x0b;
	context->signature1 = 0xface;
	if (initDecode(&context->handle)) {
		nciFreeDecodeACFR16(context_ptr);
		return 0x1770;
	}
	
	return 0;
}

static int nciFreeEncodeACFR16(void** context_ptr)
{
	NciContext* context = *((NciContext **)context_ptr);
	if (!context) {
		return 0x181a;
	}
	
	if (context->signature1 != 0xface) {
		return 0x17d4;
	}
	
	if (context->signature2 != 0x0a) {
		return 0x17de;
	}
	
	context->signature1 = 0;
	context->signature2 = 0;
	freeEncode(&context->handle);
	
	if (!context) {
		return 0;	
	}
	
	free(context);
	*context_ptr = 0;
	return 0;
}

static int nciInitEncodeACFR16(void** context_ptr)
{
	*context_ptr = malloc(44);
	if (!*context_ptr) {
		return 0x1770;
	}
	
	NciContext* context = *((NciContext **)context_ptr);
	context->signature2 = 0x0a;
	context->signature1 = 0xface;
	if (initEncode(&context->handle)) {
		nciFreeEncodeACFR16(context_ptr);
		return 0x1770;
	}
	return 0;
}

// 36c80
static int const table0[] = {
	2, 2, 2, 2,
	2, 2, 2, 2,
	2, 3, 3, 4,
	4, 5, 6, 6,
	7, 8, 9, 10,
	12, 14, 15, 0
};

// 3050
static uint32_t const band_bound_tbl[] = {
	0, 2, 4, 6,
	8, 10, 12, 14,
	16, 18, 21, 24,
	28, 32, 37, 43,
	49, 56, 64, 73,
	83, 95, 109, 124
};

// 348c
static short const gain_bit_tbl[] = {
	6, 5, 5, 5, 5, 5, 5, 5,
	5, 5, 5, 5, 5, 5, 5, 5,
	5, 5, 5, 5, 5, 5, 5, 0,
	0, 0, 0, 0, 0, 0, 0, 0
};

// 3218
static float const table1[] = {
	3134, 5342, 6870, 7792,
	8569, 9185, 9744, 10191,
	10631, 11061, 11434, 11770,
	12116, 12513, 12925, 13300,
	13674, 14027, 14352, 14716,
	15117, 15477, 15824, 16157,
	16513, 16804, 17090, 17401,
	17679, 17948, 18238, 18520,
	18764, 19078, 19381, 19640,
	19921, 20205, 20500, 20813,
	21162, 21465, 21794, 22137,
	22453, 22756, 23067, 23350,
	23636, 23926, 24227, 24521,
	24819, 25107, 25414, 25730,
	26120, 26497, 26895, 27344,
	27877, 28463, 29426, 31355
};

// 3438
static float const table2[] = {
	-11725, -9420, -7910, -6801,
	-5948, -5233, -4599, -4039,
	-3507, -3030, -2596, -2170,
	-1774, -1383, -1016, -660,
	-329, -1, 337, 696,
	1085, 1512, 1962, 2433,
	2968, 3569, 4314, 5279,
	6622, 8154, 10076, 12975
};

// *2fd8
static float const table3[] = {
	 0.0,          -0.847256005,    0.722470999,    -1.52474797,
	-0.453148007,   0.375360996,    1.47178996,     -1.98225796,
	-1.19293797,   -0.582937002,   -0.0693780035,    0.390956998,
	 0.906920016,   1.486274,       2.22154093,     -2.38878703,
	-1.80675399,   -1.41054201,    -1.07736099,     -0.799501002,
	-0.555810988,  -0.333402008,   -0.132449001,     0.0568020009,
	 0.254877001,   0.477355003,    0.738685012,     1.04430604,
	 1.39544594,    1.80987501,     2.39187598,     -2.38938308,
	-1.98846805,   -1.75140405,    -1.56431198,     -1.39221299,
	-1.216465,     -1.04694998,    -0.890510023,    -0.764558017,
	-0.645457983,  -0.52592802,    -0.405954987,    -0.302971989,
	-0.209690005,  -0.123986997,   -0.0479229987,    0.025773,
	 0.100134,      0.173718005,    0.258554012,     0.352290004,
	 0.456988007,   0.576775014,    0.700316012,     0.842552006,
	 1.00938797,    1.18213499,     1.35345602,      1.53208196,
	 1.73326194,    1.97223496,     2.39781404,     -2.5756309,
	-2.05733204,   -1.89849198,    -1.77278101,     -1.66626,
	-1.57421803,   -1.49933195,    -1.43166399,     -1.36522806,
	-1.30009902,   -1.22809303,    -1.15885794,     -1.09212506,
	-1.013574,     -0.920284986,   -0.828705013,    -0.737488985,
	-0.644775987,  -0.559094012,   -0.485713989,    -0.411031991,
	-0.345970005,  -0.285115987,   -0.234162003,    -0.187058002,
	-0.144250005,  -0.110716999,   -0.0739680007,   -0.0365610011,
	-0.00732900016, 0.0203610007,   0.0479039997,    0.0751969963,
	 0.0980999991,  0.122038998,    0.145899996,     0.169434994,
	 0.197045997,   0.225243002,    0.255686998,     0.287010014,
	 0.319709986,   0.352582991,    0.388906986,     0.433492005,
	 0.476945996,   0.520482004,    0.564453006,     0.612204015,
	 0.668592989,   0.734165013,    0.803215981,     0.878404021,
	 0.956620991,   1.03970695,     1.12937701,      1.22111595,
	 1.30802798,    1.40248001,     1.50568199,      1.62277305,
	 1.77249599,    1.94308805,     2.29039311,      0.0
};

static float const table4[] = {
	0.999981225,    0.999529421,    0.998475611,     0.996820271,
	0.994564593,    0.991709828,    0.988257587,     0.984210074,
	0.979569793,    0.974339426,    0.968522072,     0.962121427,
	0.955141187,    0.947585583,    0.939459205,     0.930767,
	0.921513975,    0.911705971,    0.901348829,     0.890448689,
	0.879012227,    0.867046177,    0.854557991,     0.841554999,
	0.828045011,    0.81403631,     0.799537301,     0.784556627,
	0.769103289,    0.753186822,    0.736816585,     0.720002472,
	0.702754676,    0.685083687,    0.666999876,     0.64851439,
	0.629638195,    0.610382795,    0.590759695,     0.570780694,
	0.550458014,    0.529803574,    0.50883007,      0.487550199,
	0.465976506,    0.444122106,    0.422000289,     0.399624199,
	0.377007395,    0.354163498,    0.331106305,     0.307849586,
	0.284407496,    0.260794103,    0.237023607,     0.213110298,
	0.189068705,    0.164913103,    0.1406582,       0.116318598,
	0.0919089988,   0.0674438998,   0.0429382995,    0.0184067003
};

static float const table5[] = {
	0.125,          0.124962397,    0.124849401,     0.124661297,
	0.124398097,    0.124059901,    0.123647101,     0.123159699,
	0.122598201,    0.121962801,    0.1212539,       0.120471999,
	0.119617499,    0.118690997,    0.117693,        0.116624102,
	0.115484901,    0.114276201,    0.112998702,     0.111653,
	0.110240199,    0.108760901,    0.107216097,     0.105606697,
	0.103933699,    0.102198102,    0.100400902,     0.0985433012,
	0.0966262966,   0.094651103,    0.0926188976,    0.0905309021,
	0.0883883014,   0.0861926004,   0.0839449018,    0.0816465989,
	0.0792991966,   0.076903902,    0.0744623989,    0.0719759986,
	0.069446303,    0.0668746978,   0.0642627999,    0.0616123006,
	0.0589246005,   0.0562013984,   0.0534444004,    0.0506552011,
	0.0478353985,   0.0449868999,   0.0421111993,    0.0392102003,
	0.0362856016,   0.0333391018,   0.0303725004,    0.0273876991,
	0.0243862998,   0.0213702004,   0.0183412991,    0.0153013002,
	0.0122520998,   0.0091955997,   0.00613350002,   0.00306769996
};

static float const table6[] = {
	-0.00613590004,-0.0306748003,  -0.0551952012,   -0.0796824023,
	-0.104121603,  -0.128498107,   -0.152797207,    -0.177004203,
	-0.201104596,  -0.225083902,   -0.248927593,    -0.272621393,
	-0.296150893,  -0.319501996,   -0.342660695,    -0.365613014,
	-0.388345003,  -0.410843194,   -0.433093786,    -0.455083609,
	-0.47679919,   -0.498227686,   -0.519356012,    -0.540171504,
	-0.560661614,  -0.580814004,   -0.600616515,    -0.620057225,
	-0.639124393,  -0.657806695,   -0.676092684,    -0.693971515,
	-0.711432219,  -0.728464425,   -0.745057821,    -0.761202395,
	-0.77688849,   -0.792106628,   -0.806847572,    -0.8211025,
	-0.834862888,  -0.848120272,   -0.860866904,    -0.873094976,
	-0.884797096,  -0.895966172,   -0.906595707,    -0.916679084,
	-0.926210225,  -0.935183525,   -0.943593502,    -0.95143503,
	-0.958703518,  -0.965394378,   -0.971503913,    -0.977028072,
	-0.981963873,  -0.986308098,   -0.990058184,    -0.993211925,
	-0.995767415,  -0.997723103,   -0.999077678,    -0.999830604
};

// 3294
static float const table7[] = {
	0.00613590004,  0.0184067003,   0.0306748003,    0.0429382995,
	0.0551952012,   0.0674438998,   0.0796824023,    0.0919089988,
	0.104121603,    0.116318598,    0.128498107,     0.1406582,
	0.152797207,    0.164913103,    0.177004203,     0.189068705,
	0.201104596,    0.213110298,    0.225083902,     0.237023607,
	0.248927593,    0.260794103,    0.272621393,     0.284407496,
	0.296150893,    0.307849586,    0.319501996,     0.331106305,
	0.342660695,    0.354163498,    0.365613014,     0.377007395,
	0.388345003,    0.399624199,    0.410843194,     0.422000289,
	0.433093786,    0.444122106,    0.455083609,     0.465976506,
	0.47679919,     0.487550199,    0.498227686,     0.50883007,
	0.519356012,    0.529803574,    0.540171504,     0.550458014,
	0.560661614,    0.570780694,    0.580814004,     0.590759695,
	0.600616515,    0.610382795,    0.620057225,     0.629638195,
	0.639124393,    0.64851439,     0.657806695,     0.666999876,
	0.676092684,    0.685083687,    0.693971515,     0.702754676,
	0.711432219,    0.720002472,    0.728464425,     0.736816585,
	0.745057821,    0.753186822,    0.761202395,     0.769103289,
	0.77688849,     0.784556627,    0.792106628,     0.799537301,
	0.806847572,    0.81403631,     0.8211025,       0.828045011,
	0.834862888,    0.841554999,    0.848120272,     0.854557991,
	0.860866904,    0.867046177,    0.873094976,     0.879012227,
	0.884797096,    0.890448689,    0.895966172,     0.901348829,
	0.906595707,    0.911705971,    0.916679084,     0.921513975,
	0.926210225,    0.930767,       0.935183525,     0.939459205,
	0.943593502,    0.947585583,    0.95143503,      0.955141187,
	0.958703518,    0.962121427,    0.965394378,     0.968522072,
	0.971503913,    0.974339426,    0.977028072,     0.979569793,
	0.981963873,    0.984210074,    0.986308098,     0.988257587,
	0.990058184,    0.991709828,    0.993211925,     0.994564593,
	0.995767415,    0.996820271,    0.997723103,     0.998475611,
	0.999077678,    0.999529421,    0.999830604,     0.999981225
};

// fffc9e00
static int16_t const table8[] = {
	28534, 25705, 25120, 25456, 10338, 21570, 20569, 10821,
	8236,  28277, 26995, 28263, 25701, 25376, 24936, 11378,
	29472, 28520, 29810, 26912, 29806, 41,    0,     0,
	0,     0,     0,     0,     0,     0,     0,     0
};

// fffc9e40
static int16_t const table9[] = {
	32767, 30840, 29127, 27594, 26214, 24966, 23831, 22795,
	21845, 20972, 20165, 19418, 18725, 18079, 17476, 16913,
	16384, 0,     0,     0,     0,     0,     0,     0,
	0,     0,     0,     0,     0,     0,     0,     0
};

// fffca020 (35fe0)
static float const table10[] = {
	0.0,            0.0122715384,   0.024541229,     0.0368072242,
	0.0490676723,   0.061320737,    0.0735645667,    0.0857973099,
	0.0980171412,   0.110222213,    0.122410677,     0.134580716,
	0.146730468,    0.158858135,    0.170961887,     0.183039889,
	0.195090324,    0.207111374,    0.219101235,     0.231058106,
	0.242980182,    0.254865646,    0.266712755,     0.27851969,
	0.290284693,    0.302005947,    0.313681751,     0.32531029,
	0.336889863,    0.348418683,    0.359895051,     0.371317178,
	0.382683426,    0.393992037,    0.405241311,     0.416429549,
	0.427555084,    0.438616246,    0.449611336,     0.460538715,
	0.471396744,    0.482183784,    0.492898196,     0.50353837,
	0.514102757,    0.524589658,    0.534997642,     0.545324981,
	0.555570245,    0.565731823,    0.575808167,     0.585797846,
	0.59569931,     0.605511069,    0.615231574,     0.624859512,
	0.634393275,    0.643831551,    0.653172851,     0.662415802,
	0.671558976,    0.680601001,    0.689540565,     0.698376238,
	0.707106769,    0.715730846,    0.724247098,     0.732654274,
	0.740951121,    0.749136388,    0.757208824,     0.765167296,
	0.773010433,    0.780737221,    0.78834641,      0.795836926,
	0.803207517,    0.81045723,     0.817584813,     0.824589312,
	0.831469595,    0.838224709,    0.84485358,      0.851355195,
	0.857728601,    0.863972843,    0.870086968,     0.876070082,
	0.881921232,    0.887639642,    0.893224299,     0.898674488,
	0.903989315,    0.909168005,    0.914209783,     0.919113874,
	0.923879504,    0.928506076,    0.932992816,     0.937339008,
	0.941544056,    0.945607305,    0.949528158,     0.953306019,
	0.956940353,    0.960430503,    0.963776052,     0.966976464,
	0.970031261,    0.972939968,    0.975702107,     0.97831738,
	0.980785251,    0.983105481,    0.985277653,     0.987301409,
	0.989176512,    0.990902662,    0.992479503,     0.993906975,
	0.99518472,     0.996312618,    0.997290432,     0.998118103,
	0.99879545,     0.999322355,    0.999698818,     0.999924719,
	1.0
};

// extracts the next byte from the bit stream
void bucb(BTYPE* bt, unsigned char* out, short const item_length)
{
	assert(item_length <= 8);
	
	uint8_t const* in = bt->stream + *bt->in_byte_offset_ptr;
	uint8_t item_end_pos = *bt->in_bit_offset_ptr + item_length;
	
	if (item_end_pos > 8) { // item spans 2 bytes
		*out = *in >> *bt->in_bit_offset_ptr;
		++in;
		if (++*bt->in_byte_offset_ptr >= bt->stream_size) {
			*bt->in_byte_offset_ptr = 0;
			in = bt->stream;
		}
		
		item_end_pos -= 8;
		uint8_t const mask = (1 << item_end_pos) - 1;
		*out |= (*in & mask) << (8 - *bt->in_bit_offset_ptr);
	} else {
		uint8_t const mask = (1 << item_length) - 1;
		*out = (*in >> *bt->in_bit_offset_ptr) & mask;
		if (item_end_pos == 8) {
			if (++*bt->in_byte_offset_ptr >= bt->stream_size) {
				*bt->in_byte_offset_ptr = 0;
			}
			item_end_pos = 0;
		}
	}
	
	*bt->in_bit_offset_ptr = item_end_pos;
}

static void bpcb(BTYPE* bt, unsigned char item, short item_length)
{
	assert(item_length <= 8);
	uint8_t* out = bt->stream + *bt->out_byte_offset_ptr;
	uint8_t item_end_pos = *bt->out_bit_offset_ptr + item_length;
	
	if (*bt->out_bit_offset_ptr == 0) {
		*out = item;
	} else {
		// 13fef0
		*out |= item << *bt->out_bit_offset_ptr;
	}
	
	if (item_end_pos > 8) {
		++out;
		if (++*bt->out_byte_offset_ptr >= bt->stream_size) {
			*bt->out_byte_offset_ptr = 0;
			out = bt->stream;
		}
		*out = item >> (8 - *bt->out_bit_offset_ptr);
		item_end_pos -= 8;
	} else {
		// 13fed0
		if (item_end_pos == 8) {
			if (++*bt->out_byte_offset_ptr >= bt->stream_size) {
				*bt->out_byte_offset_ptr = 0;
			}
			item_end_pos = 0;
		}
	}
	
	*bt->out_bit_offset_ptr = item_end_pos;
}

/**
 * Scales a value by a power of two, so that it uses the whole range of int32_t.
 * Returns the number of bits that can be saved in *shift.
 */
static void norm(long* val, short* shift)
{
	long v = *val;
	if (v == 0) {
		*shift += 31;
		return;
	} else if (v >= (1L << 30)) {
		return;
	}
	
	if (v > 0) {
		do {
			v <<= 1;
			++*shift;
		} while (v < (1L << 30));
	} else {
		long const floor = 1L << 31; // lowest possible 32bit value
		do {
			v <<= 1;
			++*shift;
		} while (v > floor + (1L << 30));
	}
	*val = v;
}

// TODO: make input const
void getD(short* input, short arg_4, short input_count, int upper_bound, short arg_10, int* out)
{
	*out = 0;
	if (input_count <= 0) {
		return;
	}
	
	int const var_1 = 1 << (arg_4 - 1);
	
	for (int i = 0; i < input_count; ++i) {
		//13f5c0
		int var_2 = input[i] - arg_10;
		if (var_2 < 0) {
			var_2 = 0;
		} else {
			var_2 = (var_2 + var_1) >> arg_4;
		}
		// 13f5d5
		// 13f5e0
		*out += std::min(var_2, upper_bound);
	}
}

void getFactor(short arg_0, short* arg_4, short* arg_8)
{
	if (arg_0 == 0) {
		*arg_4 = 0;
		*arg_8 = 0;
		return;
	}
	
	int32_t const sign = ((~(uint32_t)arg_0 >> 31) << 1) - 1; // edi
	
	int32_t abs = (int32_t)arg_0 * sign; // esi, same as abs(arg_0)
	
	// Scale abs just like norm() does.
	int shift = -1;
	while (!(abs & (1 << 15))) {
		// 13f650
		abs <<= 1;
		++shift;
	}
	abs >>= 1;
	
	*arg_8 = 27 - shift;
	
	// 13f65b
	
	int const table_val = table9[(abs - 0x3e00) >> 10];
	int32_t tmp = abs * table_val;
	tmp = 0x40000000 - tmp;
	tmp = (tmp + 0x4000) >> 15;
	tmp *= table_val;
	tmp = (tmp + 0x4000) >> 15;
	int32_t const tmp2 = tmp;
	tmp *= abs;
	tmp = 0x20000000 - tmp;
	tmp = (tmp + 0x4000) >> 15;
	tmp *= tmp2;
	tmp = (tmp + 0x2000) >> 14;
	tmp *= sign;
	
	if (tmp > 32767 && sign == 1) {
		// 13f6ec
		tmp = 32767;
	} else if (tmp < -32768 && sign == -1) {
		// 13f6e5
		tmp = -32768;
	}
	
	// 13f6da
	*arg_4 = (short)tmp;
}

static void wc(float* input, short input_count, int total_length, int* packed_sizes, int* leftover)
{
	short var_124;
	short var_122;
	short scaled_input[124];
	int bits_used;
	
	// EDI: input
	// ESI: index for input
	
	float max_input = 0;
	for (int i = 0; i < input_count; ++i) {
		// 13f740
		if (input[i] > max_input) {
			max_input = input[i];
		}
	}
	
	// 13f75b
	short max_input_scale = 0;
	{
		short excess_bits = 0;
		long val = (long)max_input;
		norm(&val, &excess_bits);
		max_input_scale = excess_bits - 16;
	}
	if (max_input_scale < 0) {
		// 13fd7b
		for (int i = 0; i < input_count; ++i) {
			// 13fdc0
			scaled_input[i] = (int)input[i] >> -max_input_scale;
		}
	} else {
		for (int i = 0; i < input_count; ++i) {
			// 13f800
			scaled_input[i] = (int)input[i] << max_input_scale;
		}
	}
	
	// 13f840
	getFactor(input_count, &var_124, &var_122);
	for (int i = 0; i < input_count; ++i) {
		// 13f870
		scaled_input[i] = ((int)scaled_input[i] * 3) >> 2; // *= 0.75
	}
	
	// 13f88f
	long scaled_input_sum = 0;
	for (int i = 0; i < input_count; ++i) {
		// 13f8b0
		scaled_input_sum += scaled_input[i];
	}
	
	max_input_scale += 11;
	// 13f8c7
	scaled_input_sum -= total_length << max_input_scale;
	int scaled_input_base = 0;
	{
		long val = scaled_input_sum - (total_length << max_input_scale);
		short shift = max_input_scale;
		norm(&val, &shift);
		scaled_input_base = val >> 16;
		scaled_input_base = ((int)scaled_input_base * (int)var_124) >> 15;
		//scaled_input_base = int16_t(scaled_input_base); // edi, sign-extend the lower 16 bits.
		
		shift = max_input_scale - (var_122 + shift - 31);
		if (shift >= 0) {
			scaled_input_base <<= shift;
		} else {
			// 13fd6a
			scaled_input_base >>= -shift;
		}
	}
	
	// var_11A is used for different purposes below.
	
	// 13f954
	getD(scaled_input, max_input_scale, input_count, 6, scaled_input_base, &bits_used);
	if (bits_used != total_length) {
		short diff = bits_used - total_length;
		short diff_scale = 0;
		if (diff <= 0) {
			// 13fd39
			for (; diff >= -16384; diff <<= 1) { // 0xc000
				// 13fd50
				++diff_scale;
			}
		} else {
			for (; diff < 16384; diff <<= 1) {
				//13f9d9
				++diff_scale;
			}
		}
	
		// 13f9e0
		int base_delta = ((int)diff * (int)var_124) >> 15; // esi
		//base_delta = int16_t(base_delta); // sign-extend the lower 16 bits.
		diff_scale = max_input_scale - (var_122 + diff_scale - 15);
		if (diff_scale >= 0) {
			base_delta <<= diff_scale;
		} else {
			// 13fd28
			base_delta >>= -diff_scale;
		}
		
		short num_revisions = 1; // var_13A
		short last_bits_used;
		short last_scaled_input_base; // var_12E
		for (;;) {
			// 13fa30
			last_bits_used = bits_used;
			last_scaled_input_base = scaled_input_base;
			scaled_input_base += base_delta;
			getD(scaled_input, max_input_scale, input_count, 6, scaled_input_base, &bits_used);
			if (++num_revisions > 19) {
				break;
			}
			if ((bits_used - total_length) * (last_bits_used - total_length) <= 0) {
				break;
			}
			
		}
		
		// 13faac
		if (bits_used != total_length) {
			short scaled_input_base_1; // var_130
			short bits_used_1; // var_134
			short bits_used_2; // var_136
			if (bits_used > total_length) {
				scaled_input_base_1 = scaled_input_base;
				scaled_input_base = last_scaled_input_base;
				bits_used_1 = bits_used;
				bits_used_2 = last_bits_used;
			} else {
				// 13fcfa
				scaled_input_base_1 = last_scaled_input_base;
				bits_used_1 = last_bits_used;
				bits_used_2 = bits_used;
			}
			
			// 13faeb
			while (bits_used != total_length && num_revisions < 20) {
				// 13fc70
				short const avg = (scaled_input_base + scaled_input_base_1) >> 1; // si
				getD(scaled_input, max_input_scale, input_count, 6, avg, &bits_used);
				++num_revisions;
				if (bits_used > total_length) {
					scaled_input_base_1 = avg;
					bits_used_1 = bits_used;
				} else {
					// 13fce9
					scaled_input_base = avg;
					bits_used_2 = bits_used;
				}
			}
			
			// 13fafe
			int const dev_1 = abs((int)bits_used_1 - total_length); // edx
			
			// 13fb12
			int const dev_2 = abs((int)bits_used_2 - total_length); // eax
			
			// 13fb26
			if (dev_1 < dev_2) {
				scaled_input_base = scaled_input_base_1;
				bits_used = bits_used_1;
			} else {
				// 13fc4f
				bits_used = bits_used_2;
			}
		}
	}
	// 13fb39
	
	for (int i = 0; i < input_count; ++i) {
		// 13fb60
		
		int tmp = (int)scaled_input[i] - scaled_input_base;
		if (tmp >= 0) {
			tmp = (tmp + (1 << (max_input_scale - 1))) >> max_input_scale;
		} else {
			// 13fc40
			tmp = 0;
		}
		// 13fb88
		// 13fb98
		packed_sizes[i] = std::min<int>(tmp, 6);
	}
	
	// 13fba7
	if (bits_used > total_length) {
		int i = 0; // esi
		int bit_count = 0; // eax
		for (; bit_count < total_length; ++i) {
			bit_count += packed_sizes[i];
		}
		// 13fbd0
		bit_count -= packed_sizes[i - 1];
		packed_sizes[i - 1] = total_length - bit_count;
		bits_used = total_length;
		for (; i < input_count; ++i) {
			packed_sizes[i] = 0;
		}
	}
	
	// 13fc18
	*leftover = total_length - bits_used;
}

static void HarXfmHelper(float* data, int half_len)
{
	int const len = half_len << 1; // edi
	
	int j = 1; // ecx
	for (int i = 1; i < len; i += 2) {
		// 140020
		if (i < j) {
			std::swap(data[i], data[j]);
			std::swap(data[i - 1], data[j - 1]);
		}
		// 140044
		int x = half_len; // arith
		while (x > 1 && x < j) { // signed
			// 140060
			j -= x;
			x >>= 1;
		}
		// 14006d
		j += x;
	}
}

static void HarXfm(float* out, int half_len_log2)
{
	int const half_len = 1 << half_len_log2;
	assert(out);
	assert(half_len_log2 > 1);
	assert(half_len >= 4);
	assert(half_len <= 256);
	
	HarXfmHelper(out, half_len);
	
	float* p = out; // edx
	for (int i = half_len >> 1; i > 0; --i, p += 4) {
		// 1400e0
		float const p0 = p[0];
		float const p1 = p[1];
		float const p2 = p[2];
		float const p3 = p[3];
		p[0] = p0 + p2;
		p[1] = p1 + p3;
		p[2] = p0 - p2;
		p[3] = p1 - p3;
	}
	// 140117
	p = out;
	for (int i = half_len >> 2; i > 0; --i, p += 8) {
		// 140130
		float const p0 = p[0];
		float const p1 = p[1];
		float const p2 = p[2];
		float const p3 = p[3];
		float const p4 = p[4];
		float const p5 = p[5];
		float const p6 = p[6];
		float const p7 = p[7];
		p[0] = p0 + p4;
		p[1] = p1 + p5;
		p[2] = p2 + p7;
		p[3] = p3 - p6;
		p[4] = p0 - p4;
		p[5] = p1 - p5;
		p[6] = p2 - p7;
		p[7] = p3 + p6;
	}
	
	int i = 0;
	int x = half_len >> 3;
	int y = 64;
	int z = 4;
	for (int idx1 = half_len_log2 - 2; idx1 > 0; --idx1, z <<= 1, y >>= 1, x >>= 1) {
		// 1401c9
		assert(x > 0);
		p = out;
		for (int idx2 = x; idx2 != 0; --idx2, p += z << 1) {
			// 1401dd
			for (int idx3 = z >> 1; idx3 > 0; --idx3, p += 2, i += y) {
				// 1401f0
				float* q = &p[z << 1];
				float const p0 = p[0];
				float const p1 = p[1];
				float const q0 = q[0];
				float const q1 = q[1];
				//14020a
				q[0] = p0 - (q0 * table10[128 - i] + q1 * table10[i]);
				p[0] = p0 + q0 * table10[128 - i] + q1 * table10[i];
				q[1] = p1 + q0 * table10[i] - q1 * table10[128 - i];
				p[1] = p1 - (q0 * table10[i] - q1 * table10[128 - i]);
				assert(i <= 0x80);
			}
			// 140266
			for (int idx4 = z >> 1; idx4 > 0; --idx4, p += 2, i -= y) {
				// 140280
				float* q = &p[z << 1];
				float const p0 = p[0];
				float const p1 = p[1];
				float const q0 = q[0];
				float const q1 = q[1];
				q[0] = p0 + q0 * table10[128 - i] - q1 * table10[i];
				p[0] = p0 - (q0 * table10[128 - i] - q1 * table10[i]);
				q[1] = p1 + q1 * table10[128 - i] + q0 * table10[i];
				p[1] = p1 - (q1 * table10[128 - i] + q0 * table10[i]);
				assert(i <= 0x80);
				assert(i >= 0);
			}
			// 1402f6
		}
		// 140305
	}
}

static void auxceps(float* input, int len_log2, float* out)
{
	float const pi = 3.14159274;
	int const len = 1 << len_log2;
	int const half_len_m1 = (len >> 1) - 1;
	int const quarter_len = len >> 2;
	
	// These can be considered as table definitions.
	assert(fabs(table4[half_len_m1] - cos((half_len_m1 + 0.25) * pi / len)) < 1E-05);
	assert(fabs(table5[half_len_m1] - cos(half_len_m1 * pi / len) * sqrt(2.0 / len)) < 1E-05);
	assert(fabs(table6[half_len_m1] + sin((half_len_m1 + 0.25) * pi / len)) < 1E-05);
	
	for (int i = 0; i < quarter_len; ++i) {
		// 140610
		int const pow2 = i << 1;
		int const j = len - 1 - pow2;
		int const k = j - 1;
		
		double const inp_pow2 = input[pow2];
		double const inp_pow2_1 = input[pow2 + 1];
		double const inp_j = input[j];
		double const inp_k = input[k];
		
		out[pow2] = table4[i] * inp_pow2 - table6[i] * inp_j;
		out[pow2 + 1] = inp_j * table4[i] + inp_pow2 * table6[i];
		
		out[k] = table4[half_len_m1 - i] * inp_k - table6[half_len_m1 - i] * inp_pow2_1;
		out[j] = inp_pow2_1 * table4[half_len_m1 - i] + inp_k * table6[half_len_m1 - i];
	}
	
	HarXfm(out, len_log2 - 1);
	
	double const last_out = out[len - 1];
	double const pre_last_out = out[len - 2];
	
	out[0] = table5[0] * out[0];
	out[len - 1] = out[1] * -table5[0];
	
	out[len - 2] = table5[half_len_m1] * out[len - 2] + table5[1] * last_out;
	out[1] = pre_last_out * table5[1] - last_out * table5[half_len_m1];
	
	float* p_out = &out[len - 3];
	float const* p_tbl = &table5[half_len_m1];
	int j = 3;
	for (int i = 1; i < quarter_len; ++i, --p_tbl, p_out -= 2, j += 2) {
		// 140740
		double const old_out_a = *p_out;
		double const old_out_b = p_out[-1];
		double const old_out_c = out[j];
		double const old_out_d = out[j - 1];
		
		out[j - 1] = *p_tbl * old_out_c + table5[(j - 1) / 2] * old_out_d;
		out[j] = old_out_b * table5[(j + 1) / 2] - old_out_a * p_tbl[-1];
		*p_out = old_out_d * *p_tbl - old_out_c * table5[(j - 1) / 2],
		p_out[-1] = table5[(j + 1) / 2] * old_out_a + p_tbl[-1] * old_out_b;
		
	}
}

static void iTransfm(float* state, float* input, int len_log2, float const* table, float* out)
{
	// TODO: make input arg const
	assert(state);
	assert(input);
	assert(table);
	assert(out);
	assert(len_log2 > 0);
	int const len = 1 << len_log2;
	int const quarter_len = len >> 2;
	int y = len - 1;
	int x = len >> 1;
	int j = x - 1;
	int i = 0;
	
	/*
	i-->      <--j
	----------------------------
	              x-->      <--y
	*/
	
	auxceps(input, len_log2, out);
	
	for (; i < quarter_len; ++i, --j, ++x, --y) {
		//130ac0
		double const state_i = state[i];
		double const state_j = state[j];
		double const out_x = out[x];
		double const out_y = out[y];
		
		state[i] = -out[j];
		state[j] = -out[i];
		
		out[i] = state_i * table[y] + out_x * table[i];
		out[j] = state_j * table[x] + out_y * table[j];
		
		out[x] = table[x] * -out_y + table[j] * state_j;
		out[y] = table[y] * -out_x + table[i] * state_i;
	}
}

static void fTransfm(float* state, float* input, int len_log2, float const* table, float* out)
{
	// TODO: make input arg const
	assert(state);
	assert(input);
	assert(table);
	assert(out);
	assert(len_log2 > 0);
	int const len = 1 << len_log2;
	int const quarter_len = len >> 2;
	int y = len - 1;
	int x = len >> 1;
	int j = x - 1;
	int i = 0;
	
	/*
	i-->      <--j
	----------------------------
	              x-->      <--y
	*/
	
	for (; i < quarter_len; ++i, ++x, --y, --j) {
		// 1408f0
		out[x] = state[i];
		out[y] = state[j];
		
		out[i] = -input[j] * table[x] - input[x] * table[j];
		out[j] = -input[y] * table[i] - input[i] * table[y];
		
		state[i] = input[i] * table[i] - input[y] * table[y];
		state[j] = input[j] * table[j] - input[x] * table[x];
	}
	// 140986
	auxceps(out, len_log2, out); //(float*, int, float*)
}

static void decode(void* handle, float* out_samples, unsigned char* input,
	unsigned short* byte_offset_ptr, uint8_t* bit_offset_ptr, uint16_t input_size)
{
	uint8_t unpacked_input[124];
	float var_808[128];
	float var_608[124];
	float var_418[124];
	BTYPE bt; // ffc8
	bt.stream = input;
	bt.stream_size = input_size;
	bt.in_byte_offset_ptr = byte_offset_ptr;
	bt.in_bit_offset_ptr = bit_offset_ptr;
	
	bucb(&bt, &unpacked_input[0], gain_bit_tbl[0]);
	var_808[0] = table1[unpacked_input[0]]; // 13ea9a
	
	for (int i = 1; i < 23; ++i) {
		bucb(&bt, &unpacked_input[i], gain_bit_tbl[i]); // 13ead0
		var_808[i] = var_808[i - 1] + table2[unpacked_input[i]];
		// 13eae7, 13eaeb
	};
	
	for (int i = 0; i < 23; ++i) {
		float const pow_res = pow(
			2.0, var_808[i] * 0.5 * 0.0009765625
		);
	
		uint32_t bound = band_bound_tbl[i]; // 13eb3a
		uint32_t const next_bound = band_bound_tbl[i + 1];
		for (; bound < next_bound; ++bound) {
			var_418[bound] = var_808[i];
			var_608[bound] = pow_res; // 13eb70
		}
	}
	
	// 13eb86
	
	int leftover;
	int packed_byte_sizes[124];
	wc(var_418, 124, 198, packed_byte_sizes, &leftover);
	
	float* p_out = out_samples;
	for (int block = 0; block < 2; ++block, p_out += 128) {
		for (int i = 0; i < 124; ++i) {
			int const packed_size = packed_byte_sizes[i];
			float val = var_608[i];
			if (packed_size > 0) {
				int const pow2 = 1 << packed_size;
				bucb(&bt, &unpacked_input[i], packed_size);
				val *= table3[pow2 - 1 + unpacked_input[i]];
			} else {
				double const rnd_u32 = rand() * (4294967296.0 / ((double)RAND_MAX + 1));
				// if (FlashRand() < 16384.0)
				if (rnd_u32 < (1<<30) + (1<<14)) {
				    val *= -0.707099974;
				} else {
				    val *= 0.707099974;
				}
			}
			var_808[i] = val;
		}
		
		for (int i = 124; i < 128; ++i) {
			var_808[i] = 0; // 13ec2f
		}
		
		for (int i = leftover; i > 0; i -= 8) {
			uint8_t throwaway;
			if (i > 8) {
				bucb(&bt, &throwaway, 8); // 13ec60
			} else {
				bucb(&bt, &throwaway, i);
				break;
			}
		}
		
		iTransfm((float*)handle, var_808, 7, table7, p_out);
	}
}

static void float_to_short_samples(
	float* float_samples, short* short_samples, int num_samples)
{
	if (num_samples <= 0) {
		return;
	}
	
	for (int i = 0; i < num_samples; ++i) {
		float const sample = float_samples[i];
		if (sample >= 32767.0) {
			short_samples[i] = 32767;
		} else if (sample <= -32768.0) {
			short_samples[i] = -32768;
		} else {
			short_samples[i] = (short)sample;
		}
	}
}

static int nciDecodeACFR16(void* context, NCI_CODEC_IO_BLOCK* block)
{
	NciContext* ctx = (NciContext*)context;
	if (!ctx) {
		return 0x17d4;
	}
	
	if(ctx->signature1 != 0xface) {
		return 0x17d4;
	}
	
	if (ctx->signature2 != 0x0b) {
		return 0x17e3;
	}
	
	if (!block) {
		return 0x181a;
	}
	
	if (!block->compressed_bytes) {
		return 0x17f2;
	}
	
	if (!block->uncompressed_samples) {
		return 0x17ed;
	}
	
	if (block->decomp_byte_offset >= block->num_bytes) {
		return 0x1801;
	}
	
	if (block->compr_byte_offset >= block->num_bytes) { // 13e712
		return 0x1806;
	}
	
	if (block->decomp_bit_offset > 7) {
		return 0x17f7;
	}
	
	if (block->compr_bit_offset > 7) {
		return 0x17fc;
	}
	
	float fbuf[256];
	decode(
		ctx->handle, fbuf, block->compressed_bytes,
		&block->decomp_byte_offset, &block->decomp_bit_offset, block->num_bytes
	);
	
	float_to_short_samples(fbuf, block->uncompressed_samples, 256);
	block->num_samples = 256;
	return 0;
}

/**
 * Finds the value closest to target and returns its index.
 */
static int findClosest(float target, float const* values, int count)
{
	assert(values);
	assert(count > 0);
	float min_distance = fabs(target - values[0]);
	int min_idx = 0;
	for (int i = 1; i < count; ++i) {
		// 13f3e0
		float const distance = fabs(target - values[i]);
		if (distance < min_distance) {
			min_distance = distance;
			min_idx = i;
		}
	}
	return min_idx;
}

/**
 * Find the value closest to target and returns its index.
 * Assumes that the values are ordered from the lowest to the highest.
 */
static int findClosestOrdered(float target, float const* values, int count)
{
	assert(values);
	assert(count > 1);
	int begin = 0;
	int end = count;
	do {
		// 13f2b0
		int const middle = (begin + end) >> 1;
		if (target > values[middle]) {
			begin = middle;
		} else {
			// 13f360
			end = middle;
		}
		// 13f2cf
	} while (end - begin > 1);
	
	assert(end > 0);
	assert(end <= count);
	assert(begin >= 0); // signed
	assert(begin < count);
	
	if (end != count) {
		// 13f2f9
		if (target - values[begin] > values[end] - target) {
			begin = end;
		}
	}
	return begin;
}

static void encode(void* state, float* samples, unsigned char* out,
	unsigned short* byte_offset_ptr, unsigned char* bit_offset_ptr, unsigned short out_len)
{
	float var_8d8[256];
	float var_2e8[23];
	float var_268[23];
	float var_4d8[124];
	float var_208[124];
	BTYPE bt;
	
	bt.stream = out;
	bt.stream_size = out_len;
	bt.out_byte_offset_ptr = byte_offset_ptr;
	bt.out_bit_offset_ptr = bit_offset_ptr;
	
	fTransfm((float*)state, samples, 7, table7, var_8d8);
	fTransfm((float*)state, samples + 128, 7, table7, var_8d8 + 128);
	
	for (int i = 0; i < 23; ++i) {
		// 13ee4e
		int bound = band_bound_tbl[i]; // esi
		int const next_bound = band_bound_tbl[i + 1]; // eax
		double acc = 0.0;
		for (; bound < next_bound; ++bound) {
			// 13ee70
			double const a = var_8d8[bound];
			double const b = var_8d8[bound + 128];
			acc += a*a + b*b;
		}
		// 13ee8f
		double const tmp = std::max<double>(1.0, acc / (table0[i] << 1));
		// 13eeb6
		var_2e8[i] = floor(log(tmp) * (1.44269502 * 1024.0) + 0.5);
	};
	
	int idx = findClosest(var_2e8[0], table1, 64);
	var_268[0] = table1[idx];
	bpcb(&bt, idx, gain_bit_tbl[0]);
	for (int i = 1; i < 23; ++i) {
		idx = findClosest(var_2e8[i] - var_268[i - 1], table2, 32);
		var_268[i] = var_268[i - 1] + table2[idx];
		bpcb(&bt, idx, gain_bit_tbl[i]);
	}
	for (int i = 0; i < 23; ++i) {
		// 13eff6
		var_2e8[i] = 1.0 / pow(2.0, var_268[i] * (0.5 * 0.0009765625));
	}
	for (int i = 0; i < 23; ++i) {
		// 13f03c
		int bound = band_bound_tbl[i]; // esi
		int const next_bound = band_bound_tbl[i + 1]; // eax
		for (; bound < next_bound; ++bound) {
			// 13f060
			var_4d8[bound] = var_268[i];
			var_208[bound] = var_2e8[i];
		}
		// 13f080
	}
	
	int leftover;
	int packed_byte_sizes[124];
	wc(var_4d8, 124, 198, packed_byte_sizes, &leftover);
	float* p = var_8d8;
	for (int block = 0; block < 2; ++block, p += 128) {
		// 13f0be
		for (int i = 0; i < 124; ++i) {
			// 13f0c0
			int const packed_size = packed_byte_sizes[i]; // ecx
			if (packed_size > 0) {
				int const pow2 = 1 << packed_size;
				idx = findClosestOrdered(var_208[i] * p[i], &table3[pow2 - 1], pow2);
				bpcb(&bt, idx, packed_size); // BTYPE*, uchar, short
			}
			// 13f130
		}
		for (int i = leftover; i > 0; i -= 8) {
			if (i > 8) {
				bpcb(&bt, 0, 8);
			} else {
				bpcb(&bt, 0, i);
				break;
			}
		}
		// 13f16a
	}
}

static int nciEncodeACFR16(void* context, NCI_CODEC_IO_BLOCK* block)
{
	NciContext* ctx = (NciContext*)context; // ecx
	if (!ctx) {
		return 0x17d4;
	}
	if (ctx->signature1 != 0xface) {
		return 0x17d4;
	}
	if (ctx->signature2 != 0x0a) {
		return 0x17de;
	}
	if (!block) {
		return 0x181a;
	}
	if (!block->compressed_bytes) {
		return 0x17f2;
	}
	if (!block->uncompressed_samples) {
		return 0x17ed;
	}
	if (block->decomp_byte_offset >= block->num_bytes) {
		return 0x1801;
	}
	if (block->compr_byte_offset >= block->num_bytes) {
		return 0x1806;
	}
	if (block->decomp_bit_offset > 7) {
		return 0x17f7;
	}
	if (block->compr_bit_offset > 7) {
		return 0x17fc;
	}
	float samples[256];
	for (int i = 0; i < 256; ++i) {
		samples[i] = block->uncompressed_samples[i];
	}
		
	encode(
		ctx->handle, samples, block->compressed_bytes,
		&block->compr_byte_offset, &block->compr_bit_offset, block->num_bytes
	);
	
	return 0;
}


static uint32_t readBytesBE(uint8_t const* from, int num)
{
	uint32_t val = 0;
	for (; num > 0; --num, ++from) {
		val <<= 8;
		val |= *from;
	}
	return val;
}

class Decompressor
{
public:
	Decompressor();
	
	~Decompressor();
	
	void processFLV(char const* fname);
private:
	void processInput(uint8_t const* input, size_t size, bool eof);
	
	void processOutputBlock();
	
	uint8_t m_inputBuffer[64];
	short m_outputBuffer[256];
	size_t m_inputBufFreeSpace;
	size_t m_outputBufFreeSpace;
	SNDFILE* out;
};

Decompressor::Decompressor()
:	m_inputBufFreeSpace(64),
	m_outputBufFreeSpace(256)
{
	SF_INFO info;
	memset(&info, 0, sizeof(info));
	info.samplerate = 8000;
	info.channels = 1;
	info.format = SF_FORMAT_WAV|SF_FORMAT_PCM_16;
	out = sf_open("test2.wav", SFM_WRITE, &info);
	if (!out) {
		throw std::runtime_error("Unable to create test2.wav");
	}
}

Decompressor::~Decompressor()
{
	sf_close(out);
}

void
Decompressor::processFLV(char const* fname)
{
	FILE* fp = fopen(fname, "rb");
	if (!fp) {
		throw std::runtime_error("Unable to open "+std::string(fname));
	}
	
	char header[9];
	fread(header, 1, 9, fp);
	
	FlvTag tag;
	uint8_t tag_header[FLV_TAG_SIZE];
	size_t bytes_to_ignore = 64;
	while (fread(tag_header, 1, FLV_TAG_SIZE, fp) == FLV_TAG_SIZE) {
		tag.prev_tag_len = readBytesBE(tag_header, 4);
		tag.type = tag_header[4];
		tag.data_len = readBytesBE(tag_header + 5, 3);
		tag.timestamp = readBytesBE(tag_header + 8, 3);
		std::cout << "timestamp: " << tag.timestamp << std::endl;
		
		if (tag.data_len == 0) {
			std::cerr << "tag.data_len == 0" << std::endl;
			continue;
		}
		
		std::vector<uint8_t> data(tag.data_len);
		fread(&data[0], 1, tag.data_len, fp);
		
		if (tag.type != AUDIO_TAG) {
			continue;
		}
		int const format = data[0];
		std::cout << "format = " << format << std::endl;
		
		uint8_t const* input = &data[1];
		size_t input_size = data.size() - 1;
		if (bytes_to_ignore) {
			size_t const ignore = std::min(bytes_to_ignore, input_size);
			input += ignore;
			input_size -= ignore;
			bytes_to_ignore -= ignore;
		}
		if (input_size > 0) {
			processInput(input, input_size, false);
		}
	}
	processInput(0, 0, true);
}

void
Decompressor::processInput(uint8_t const* input, size_t size, bool eof)
{
	NCI_CODEC_IO_BLOCK block;
	
	//fwrite(input, 1, size, stdout);
	
	while (size || eof) {
		if (size) {
			size_t const chunk = std::min(m_inputBufFreeSpace, size);
			memcpy(m_inputBuffer + 64 - m_inputBufFreeSpace, input, chunk);
			m_inputBufFreeSpace -= chunk;
			if (m_inputBufFreeSpace == 0) {
				m_inputBufFreeSpace = 64;
			}
			input += chunk;
			size -= chunk;
		}
		
		if (m_inputBufFreeSpace == 64 || eof) {
			if (m_inputBufFreeSpace != 64) {
				memset(m_inputBuffer + 64 - m_inputBufFreeSpace, 0, m_inputBufFreeSpace);
			}
			memset(m_outputBuffer, 0, sizeof(m_outputBuffer));
			memset(&block, 0, sizeof(block));
			block.uncompressed_samples = m_outputBuffer;
			block.num_samples = 256;
			block.compressed_bytes = m_inputBuffer;
			block.num_bytes = 64;
			nciDecodeACFR16(decoder, &block);
			processOutputBlock();
			if (eof) {
				break;
			}
		}
	}
}

void
Decompressor::processOutputBlock()
{
	//fwrite(m_outputBuffer, 1, sizeof(m_outputBuffer), stderr);
	sf_write_short(out, m_outputBuffer, 256);
}

class Compressor
{
public:
	Compressor();
	
	~Compressor();
	
	void processWAV();
private:
	void processOutputBlock();
	
	short m_inputBuffer[256];
	uint8_t m_outputBuffer[64];
	size_t m_inputBufFreeSpace;
	size_t m_outputBufFreeSpace;
	FILE* out;
	
};

Compressor::Compressor()
:	m_inputBufFreeSpace(256),
	m_outputBufFreeSpace(64)
{
	out = fopen("test2.flv", "wb");
	if (!out) {
		throw std::runtime_error("Could not open test2.flv for writing");
	}
	
	unsigned char header[9];
	header[0] = 'F';
	header[1] = 'L';
	header[2] = 'V';
	header[3] = 1;
	header[4] = 4;
	header[5] = 0;
	header[6] = 0;
	header[7] = 0;
	header[8] = 9;
	fwrite(header, 1, 9, out);
	
	unsigned char tag[15];
	tag[0] = 0;
	tag[1] = 0;
	tag[2] = 0;
	tag[3] = 0;
	tag[4] = AUDIO_TAG;
	tag[5] = 0; // \
	tag[6] = 0; // } body length + 1: write after we are done
	tag[7] = 0; // /
	tag[8] = 0;
	tag[9] = 0;
	tag[10] = 0;
	tag[11] = 0;
	tag[12] = 0;
	tag[13] = 0;
	tag[14] = 0;
	fwrite(tag, 1, 15, out);
	
	unsigned char format = 2 | (5 << 4);
	fwrite(&format, 1, 1, out);
}

Compressor::~Compressor()
{
	size_t const file_length = ftell(out);
	size_t const body_length = file_length - 9 - 15;
	unsigned char bl[3];
	bl[0] = (body_length >> 16) & 0xff;
	bl[1] = (body_length >> 8) & 0xff;
	bl[2] = (body_length) & 0xff;
	fseek(out, 14, SEEK_SET);
	fwrite(bl, 1, 3, out);
	fclose(out);
}

void
Compressor::processWAV()
{
	SF_INFO info;
	memset(&info, 0, sizeof(info));
	SNDFILE* in = sf_open("test3.wav", SFM_READ, &info);
	if (!in) {
		throw std::runtime_error("Unable to open test3.wav");
	}
	
	NCI_CODEC_IO_BLOCK block;
	
	int read = 0;
	for (;;) {
		read = sf_read_short(in, m_inputBuffer, 256);
		if (read < 256) {
			m_inputBufFreeSpace = 256 - read;
			break;
		}
		//fwrite(m_inputBuffer, 1, 256*sizeof(m_inputBuffer[0]), stdout);
		memset(&block, 0, sizeof(block));
		block.uncompressed_samples = m_inputBuffer;
		block.compressed_bytes = m_outputBuffer;
		block.num_bytes = 64;
		nciEncodeACFR16(encoder, &block);
		processOutputBlock();
	}
	
	if (m_inputBufFreeSpace < 256) {
		memset(&m_inputBuffer[read], 0, m_inputBufFreeSpace*sizeof(m_inputBuffer[0]));
		//fwrite(m_inputBuffer, 1, 256*sizeof(m_inputBuffer[0]), stdout);
		memset(&block, 0, sizeof(block));
		block.uncompressed_samples = m_inputBuffer;
		block.compressed_bytes = m_outputBuffer;
		block.num_bytes = 64;
		nciEncodeACFR16(encoder, &block);
		processOutputBlock();
	}
	
	sf_close(in);
}

void
Compressor::processOutputBlock()
{
	fwrite(m_outputBuffer, 1, 64, out);
	//fwrite(m_outputBuffer, 1, 64, stderr);
}

int main(int argc, char** argv)
{
	if (argc == 1) {
		nciInitEncodeACFR16(&encoder);
		
		try {
			Compressor compressor;
			compressor.processWAV();
		} catch (std::exception const& e) {
			std::cout << e.what() << std::endl;
		}
		
		nciFreeEncodeACFR16(&encoder);
	} else {
		nciInitDecodeACFR16(&decoder);
		
		try {
			Decompressor decompressor;
			decompressor.processFLV(argv[1]);
		} catch (std::exception const& e) {
			std::cout << e.what() << std::endl;
		}
		
		nciFreeDecodeACFR16(&decoder);
	}
}
